unit Field;
interface

const MARKED=$40000000;
      UNMARK=$BFFFFFFF;

type TField=class
            private
              Data:array of integer;
              FWidth,FHeight,FQ:integer;
              FHaveMarks:boolean;
              UndoBuf:array of byte;
              UndoPos:integer;
              function ReadField(X,Y:integer):integer;
              procedure WriteField(X,Y:integer; Value:integer);
              procedure SetHaveMarks(Value:boolean);
              procedure Recall;
            public
              property Field[X,Y:integer]:integer read ReadField write WriteField; default;
              property Width:integer read FWidth;
              property Height:integer read FHeight;
              property HaveMarks:boolean read FHaveMarks write SetHaveMarks;
              property UndoCount:integer read UndoPos;
              // Basics
              constructor Create;
              destructor Destroy; override;
              procedure Init(Width,Height:integer);
              procedure InitRandomly(Width,Height,Colors:integer);
              // Markierungen setzen und lschen, Zge durchfhren usw.
              procedure RemoveMarks;
              procedure RemoveMarked;
              function MarkAdjacent(X,Y:integer):integer;
              function MarkFalling:boolean;
              procedure FallMarked;
              function MarkSliding:boolean;
              procedure SlideMarked;
              function Move(X,Y:integer):boolean;
              // Undo-Verwaltung
              procedure AddUndo;
              function CanUndo:boolean;
              function Undo:boolean;
              function CanRedo:boolean;
              function Redo:boolean;
            end;

implementation

constructor TField.Create;
begin
  inherited Create;
  SetLength(Data,0);
  SetLength(UndoBuf,0); UndoPos:=-1;
  FWidth:=0; FHeight:=0; FHaveMarks:=false;
end;

destructor TField.Destroy;
begin
  Finalize(UndoBuf);
  Finalize(Data);
  inherited Destroy;
end;

function TField.ReadField(X,Y:integer):integer;
begin
  if (X<0) OR (Y<0) OR (X>=FWidth) OR (Y>=FHeight)
    then Result:=-1
    else Result:=Data[Y*FWidth+X];
end;

procedure TField.WriteField(X,Y:integer; Value:integer);
begin
  if (X<0) OR (Y<0) OR (X>=FWidth) OR (Y>=FHeight) then exit;
  Data[Y*FWidth+X]:=Value;
end;

procedure TField.Init(Width,Height:integer);
begin
  FWidth:=Width; FHeight:=Height;
  FQ:=Width*Height;
  FHaveMarks:=false;
  SetLength(Data,FQ);
  SetLength(UndoBuf,0); UndoPos:=-1;
  FillChar(Data[0],FQ*sizeof(integer),0);
end;

procedure TField.InitRandomly(Width,Height,Colors:integer);
var i:integer;
begin
  Init(Width,Height);
  for i:=0 to (FQ-1) do
    Data[i]:=random(Colors)+1;
  AddUndo;
end;

procedure TField.SetHaveMarks(Value:boolean);
begin
  if not Value then RemoveMarks else FHaveMarks:=true;
end;

procedure TField.RemoveMarks;
var P:^integer; i:integer;
begin
  if not FHaveMarks then exit;
  P:=@Data[0];
  for i:=FQ downto 1 do begin
    P^:=P^ AND UNMARK;
    inc(P);
  end;
  FHaveMarks:=false;
end;

procedure TField.RemoveMarked;
var P:^integer; i:integer;
begin
  if not FHaveMarks then exit;
  P:=@Data[0];
  for i:=FQ downto 1 do begin
    if (P^ AND MARKED)<>0 then P^:=MARKED;
    inc(P);
  end;
end;

function TField.MarkAdjacent(X,Y:integer):integer;
var res,BaseCol:integer;
  procedure rmark(X,Y:integer);
  var i:integer;
  begin
    i:=Field[X,Y];
    if (i<>BaseCol) OR ((i AND MARKED)<>0) then exit;
    Field[X,Y]:=i OR MARKED;
    inc(res);
    rmark(X-1,Y); rmark(X+1,Y);
    rmark(X,Y-1); rmark(X,Y+1);
  end;
begin
  res:=0; RemoveMarks;
  BaseCol:=Field[X,Y];
  if BaseCol>0 then
    rmark(X,Y);
  if res>0 then FHaveMarks:=true;
  Result:=res;
end;

const MD_FIND_HOLE=0;
      MD_FIND_SOLID=1;
      MD_OK=2;

function TField.MarkFalling:boolean;
var X,Y,SY,Mode:integer;
begin
  Result:=false; RemoveMarks;
  for X:=0 to (FWidth-1) do begin
    Mode:=MD_FIND_HOLE; SY:=-1;
    for Y:=(FHeight-1) downto 0 do
      if Mode=MD_FIND_HOLE then begin
        if Field[X,Y]=0 then Mode:=MD_FIND_SOLID;
      end else
        if Field[X,Y]<>0 then begin
          SY:=Y;
          Mode:=MD_OK;
          break;
        end;
    if Mode=MD_OK then begin
      Result:=true;
      for Y:=SY downto 0 do
        Field[X,Y]:=Field[X,Y] OR MARKED;
    end;
  end;
  FHaveMarks:=Result;
end;

procedure TField.FallMarked;
var X,Y,F:integer;
begin
  F:=-1;
  for X:=0 to (FWidth-1) do begin
    for Y:=(FHeight-1) downto 0 do begin
      F:=Field[X,Y];
      if (F AND MARKED)<>0 then
        Field[X,Y+1]:=F AND UNMARK;
    end;
    if (F AND MARKED)<>0 then
      Field[X,0]:=0;
  end;
  FHaveMarks:=false;
end;

function TField.MarkSliding:boolean;
var X,Y,SX,Mode:integer;
begin
  Result:=false; RemoveMarks;
  Mode:=MD_FIND_HOLE; SX:=-1;
  for X:=0 to (FWidth-1) do
    if Mode=MD_FIND_HOLE then begin
      if Field[X,FHeight-1]=0 then Mode:=MD_FIND_SOLID;
    end else
      if Field[X,FHeight-1]<>0 then begin
        SX:=X;
        Mode:=MD_OK;
        break;
      end;
  if Mode=MD_OK then begin
    Result:=true;
    for Y:=0 to (FHeight-1) do
      for X:=SX to (FWidth-1) do
        Field[X,Y]:=Field[X,Y] OR MARKED;
  end;
  FHaveMarks:=Result;
end;

procedure TField.SlideMarked;
var X,Y,F:integer;
begin
  F:=-1;
  for Y:=0 to (FHeight-1) do begin
    for X:=1 to (FWidth-1) do begin
      F:=Field[X,Y];
      if (F AND MARKED)<>0 then
        Field[X-1,Y]:=F AND UNMARK;
    end;
    if (F AND MARKED)<>0 then
      Field[FWidth-1,Y]:=0;
  end;
  FHaveMarks:=false;
end;

function TField.Move(X,Y:integer):boolean;
begin
  Result:=(MarkAdjacent(X,Y)>1);
  if Result then begin
    while MarkFalling do FallMarked;
    while MarkSliding do SlideMarked;
  end;
end;

procedure TField.AddUndo;
var i:integer;
    Pi:^integer; Pb:^byte;
begin
  inc(UndoPos);
  SetLength(UndoBuf,FQ*(UndoPos+1));
  Pi:=@Data[0];
  Pb:=@UndoBuf[FQ*UndoPos];
  for i:=FQ downto 1 do begin
    Pb^:=Pi^;
    inc(Pb); inc(Pi);
  end;
end;

procedure TField.Recall;
var i:integer;
    Pi:^integer; Pb:^byte;
begin
  Pi:=@Data[0];
  Pb:=@UndoBuf[FQ*UndoPos];
  for i:=FQ downto 1 do begin
    Pi^:=Pb^;
    inc(Pb); inc(Pi);
  end;
end;

function TField.CanUndo:boolean;
begin
  Result:=(UndoPos>0);
end;

function TField.Undo:boolean;
begin
  Result:=CanUndo;
  if not Result then exit;
  dec(UndoPos); Recall;
end;

function TField.CanRedo:boolean;
begin
  Result:=((UndoPos+1)<(length(UndoBuf) DIV FQ));
end;

function TField.Redo:boolean;
begin
  Result:=CanRedo;
  if not Result then exit;
  inc(UndoPos); Recall;
end;

end.

